@@ -14,7 +14,7 @@ from pymongo import DESCENDING
 
 from fireworks.core.fw_config import FWConfig
 from fireworks.utilities.fw_serializers import FWSerializable
-from fireworks.core.firework import FireWork, Launch, Workflow, FWAction
+from fireworks.core.firework import FireWork, Launch, Workflow, FWAction, Tracker
 from fireworks.utilities.fw_utilities import get_fw_logger
 
 
@@ -138,12 +138,13 @@ class LaunchPad(FWSerializable):
             self.m_logger.info('Performing maintenance on Launchpad...')
 
             self.m_logger.debug('Tracking down FIZZLED jobs...')
-            fl = self.detect_fizzled(fix=True)
+            fl, ff = self.detect_lostruns(fizzle=True)
             if fl:
                 self.m_logger.info('Detected {} FIZZLED launches: {}'.format(len(fl), fl))
+                self.m_logger.info('Detected {} FIZZLED FWs: {}'.format(len(ff), ff))
 
             self.m_logger.debug('Tracking down stuck RESERVED jobs...')
-            ur = self.detect_unreserved(fix=True)
+            ur = self.detect_unreserved(rerun=True)
             if ur:
                 self.m_logger.info('Unreserved {} RESERVED launches: {}'.format(len(ur), ur))
 
@@ -294,10 +295,12 @@ class LaunchPad(FWSerializable):
         self.launches.ensure_index('time_end')
         self.launches.ensure_index('host')
         self.launches.ensure_index('ip')
+        self.launches.ensure_index('fworker.name')
 
         self.workflows.ensure_index('name')
         self.workflows.ensure_index('created_on')
         self.workflows.ensure_index('updated_on')
+        self.workflows.ensure_index('nodes')
 
         for idx in self.user_indices:
             self.fireworks.ensure_index(idx)
@@ -414,7 +417,8 @@ class LaunchPad(FWSerializable):
             # create a launch
         # TODO: this code is duplicated with checkout_fw with minimal mods, should refactor this!!
         launch_id = self.get_new_launch_id()
-        m_launch = Launch('RESERVED', launch_dir, fworker, host, ip, launch_id=launch_id,
+        trackers = [Tracker.from_dict(f) for f in m_fw.spec['_trackers']] if '_trackers' in m_fw.spec else None
+        m_launch = Launch('RESERVED', launch_dir, fworker, host, ip, trackers=trackers, launch_id=launch_id,
                           fw_id=m_fw.fw_id)
         self._upsert_launch(m_launch)
 
@@ -435,7 +439,7 @@ class LaunchPad(FWSerializable):
         for fw in self.fireworks.find({'launches': launch_id, 'state': 'RESERVED'}, {'fw_id': 1}):
             self.fireworks.find_and_modify({'fw_id': fw['fw_id']}, {'$set': {'state': 'READY'}})
 
-    def detect_unreserved(self, expiration_secs=FWConfig().RESERVATION_EXPIRATION_SECS, fix=False):
+    def detect_unreserved(self, expiration_secs=FWConfig().RESERVATION_EXPIRATION_SECS, rerun=False):
         bad_launch_ids = []
         now_time = datetime.datetime.utcnow()
         cutoff_timestr = (now_time - datetime.timedelta(seconds=expiration_secs)).isoformat()
@@ -444,7 +448,7 @@ class LaunchPad(FWSerializable):
                                              {'launch_id': 1})
         for ld in bad_launch_data:
             bad_launch_ids.append(ld['launch_id'])
-        if fix:
+        if rerun:
             for lid in bad_launch_ids:
                 self.unreserve(lid)
         return bad_launch_ids
@@ -462,19 +466,37 @@ class LaunchPad(FWSerializable):
             wf = self.get_wf_by_fw_id(fw_id)
             self._refresh_wf(wf, fw_id)
 
-    def detect_fizzled(self, expiration_secs=FWConfig().RUN_EXPIRATION_SECS, fix=False):
+    def detect_lostruns(self, expiration_secs=FWConfig().RUN_EXPIRATION_SECS, fizzle=False, rerun=False, max_runtime=None):
         bad_launch_ids = []
+        bad_fw_ids = []
         now_time = datetime.datetime.utcnow()
         cutoff_timestr = (now_time - datetime.timedelta(seconds=expiration_secs)).isoformat()
         bad_launch_data = self.launches.find({'state': 'RUNNING', 'state_history': {
             '$elemMatch': {'state': 'RUNNING', 'updated_on': {'$lte': cutoff_timestr}}}},
-                                             {'launch_id': 1})
+                                             {'launch_id': 1, 'fw_id': 1})
         for ld in bad_launch_data:
-            bad_launch_ids.append(ld['launch_id'])
-        if fix:
+            bad_launch = False
+            if max_runtime:
+                m_l = self.get_launch_by_id(ld['launch_id'])
+                utime = m_l._get_time('RUNNING', use_update_time=True)
+                ctime = m_l._get_time('RUNNING', use_update_time=False)
+                if (utime-ctime).seconds <= max_runtime:
+                    bad_launch=True
+
+            if not max_runtime or bad_launch:
+                bad_launch_ids.append(ld['launch_id'])
+                bad_fw_ids.append(ld['fw_id'])
+        if fizzle or rerun:
             for lid in bad_launch_ids:
                 self.mark_fizzled(lid)
-        return bad_launch_ids
+        if rerun:
+            for fw_id in bad_fw_ids:
+                # only rerun if the FireWork has 'FIZZLED' state. It is possible that
+                # a different Launch for the same FireWork is 'COMPLETED'
+                if self.fireworks.find_one({"fw_id": fw_id, "state": "FIZZLED"}):
+                    self.rerun_fw(fw_id)
+
+        return bad_launch_ids, bad_fw_ids
 
     def _set_reservation_id(self, launch_id, reservation_id):
         m_launch = self.get_launch_by_id(launch_id)
@@ -506,7 +528,8 @@ class LaunchPad(FWSerializable):
 
         state_history = reserved_launch.state_history if reserved_launch else None
         l_id = reserved_launch.launch_id if reserved_launch else self.get_new_launch_id()
-        m_launch = Launch('RUNNING', launch_dir, fworker, host, ip, state_history=state_history,
+        trackers = [Tracker.from_dict(f) for f in m_fw.spec['_trackers']] if '_trackers' in m_fw.spec else None
+        m_launch = Launch('RUNNING', launch_dir, fworker, host, ip, trackers=trackers, state_history=state_history,
                           launch_id=l_id,
                           fw_id=m_fw.fw_id)
 
@@ -559,14 +582,18 @@ class LaunchPad(FWSerializable):
         for fw in self.fireworks.find({'launches': launch_id}, {'fw_id': 1}):
             fw_id = fw['fw_id']
             self._refresh_wf(self.get_wf_by_fw_id(fw_id), fw_id)
-
-        return m_launch
+        # change return type to dict to make return type seriazlizable to
+        # support job packing
+        return m_launch.to_dict()
 
     def ping_launch(self, launch_id, ptime=None):
         m_launch = self.get_launch_by_id(launch_id)
+        for tracker in m_launch.trackers:
+            tracker.track_file(m_launch.launch_dir)
+
         m_launch.touch_history(ptime)
-        self.launches.find_and_modify({'launch_id': launch_id, 'state': 'RUNNING'},
-            {'$set':{'state_history':m_launch.to_db_dict()['state_history']}})
+        self.launches.update({'launch_id': launch_id, 'state': 'RUNNING'},
+            {'$set':{'state_history':m_launch.to_db_dict()['state_history'], 'trackers': [t.to_dict() for t in m_launch.trackers]}})
 
     def get_new_fw_id(self):
         """
@@ -656,6 +683,9 @@ class LaunchPad(FWSerializable):
         # TODO: this no longer needs to be its own function (much was removed)
         self.launches.find_and_modify({'launch_id': m_launch.launch_id}, m_launch.to_db_dict(), upsert=True)
 
+    def set_priority(self, fw_id, priority):
+        self.fireworks.find_and_modify({"fw_id": fw_id}, {'$set': {'spec._priority': priority}})
+
     def get_logdir(self):
         # AJ: This is needed for job packing due to Proxy objects not being fully featured...
         return self.logdir
@@ -696,7 +726,8 @@ class LaunchPad(FWSerializable):
                 if 'fwaction' in offline_data:
                     fwaction = FWAction.from_dict(offline_data['fwaction'])
                     state = offline_data['state']
-                    m_launch = self.complete_launch(launch_id, fwaction, state)
+                    m_launch = Launch.from_dict(
+                        self.complete_launch(launch_id, fwaction, state))
                     for s in m_launch.state_history:
                         if s['state'] == offline_data['state']:
                             s['created_on'] = datetime.datetime.strptime(offline_data['completed_on'], "%Y-%m-%dT%H:%M:%S.%f")
@@ -712,4 +743,14 @@ class LaunchPad(FWSerializable):
             return m_launch.fw_id
 
     def forget_offline(self, fw_id):
-        self.offline_runs.update({"fw_id": fw_id}, {"$set": {"deprecated":True}})
\ No newline at end of file
+        self.offline_runs.update({"fw_id": fw_id}, {"$set": {"deprecated":True}})
+
+    def get_tracker_data(self, fw_id):
+        data = []
+        for l in self.launches.find({'fw_id': fw_id}, {'trackers': 1, 'launch_id': 1}):
+            if 'trackers' in l:  # backwards compatibility
+                trackers = [Tracker.from_dict(t) for t in l['trackers']]
+                data.append({'launch_id': l['launch_id'], 'trackers': trackers})
+
+        return data
+
